#include "pch.h"
#include "./Intel_2019//ippbase.h"
#include "./Intel_2019/ipptypes.h"
#include "./Intel_2019/ippi.h"
#include <windows.h>
#include <stdio.h>
#include <cmath>

typedef unsigned char BYTE;

typedef IppStatus (*PFippiResizeGetSize_8u) (
	IppiSize srcSize, IppiSize dstSize,
	IppiInterpolationType interpolation, Ipp32u antialiasing,
	int* pSpecSize, int* pInitBufSize);

typedef IppStatus (*PFippiResizeCubicInit_8u) (
	IppiSize srcSize, IppiSize dstSize, Ipp32f valueB, Ipp32f valueC,
	IppiResizeSpec_32f* pSpec, Ipp8u* pInitBuf);

typedef IppStatus (*PFippiResizeSuperInit_8u) (
	IppiSize srcSize, IppiSize dstSize,
	IppiResizeSpec_32f* pSpec);

typedef IppStatus(*PFippiResizeGetBufferSize_8u) (
	const IppiResizeSpec_32f* pSpec,
	IppiSize dstSize, Ipp32u numChannels,
	int* pBufSize);

typedef Ipp8u* (*PFippsMalloc_8u) (int len);
typedef void(*PFippsFree) (void* ptr);

typedef IppStatus (*PFippiResizeCubic_8u_C3R) (
	const Ipp8u* pSrc, Ipp32s srcStep,
	Ipp8u* pDst, Ipp32s dstStep, IppiPoint dstOffset, IppiSize dstSize,
	IppiBorderType border, const Ipp8u* pBorderValue,
	const IppiResizeSpec_32f* pSpec, Ipp8u* pBuffer);

typedef IppStatus(*PFippiResizeSuper_8u_C3R) (
	const Ipp8u* pSrc, Ipp32s srcStep,
	Ipp8u* pDst, Ipp32s dstStep, IppiPoint dstOffset, IppiSize dstSize,
	const IppiResizeSpec_32f* pSpec, Ipp8u* pBuffer);

typedef IppStatus(*PFippiResizeGetSrcOffset_8u) (
	const IppiResizeSpec_32f* pSpec,
	IppiPoint dstOffset, IppiPoint* srcOffset);

PFippiResizeGetSize_8u pfippiResizeGetSize_8u = nullptr;
PFippiResizeCubicInit_8u pfippiResizeCubicInit_8u = nullptr;
PFippiResizeSuperInit_8u pfippiResizeSuperInit_8u = nullptr;
PFippiResizeCubic_8u_C3R pfippiResizeCubic_8u_C3R = nullptr;
PFippiResizeSuper_8u_C3R pfippiResizeSuper_8u_C3R = nullptr;
PFippiResizeGetBufferSize_8u pfippiResizeGetBufferSize_8u = nullptr;
PFippsMalloc_8u pfippsMalloc_8u = nullptr;
PFippsFree pfippsFree = nullptr;
PFippiResizeGetSrcOffset_8u pfippiResizeGetSrcOffset_8u = nullptr;

bool InitIntel2019()
{
	HMODULE hModuleA = GetModuleHandle("ippi.dll");
	if (!hModuleA)
	{
		hModuleA = LoadLibrary("ippi.dll");
		if (!hModuleA)
		{
			MessageBox(0, "Fail to load ippi.dll", 0, MB_OK);
			return false;
		}
	}

	pfippiResizeGetSize_8u = (PFippiResizeGetSize_8u)GetProcAddress(hModuleA, "ippiResizeGetSize_8u");
	pfippiResizeCubicInit_8u = (PFippiResizeCubicInit_8u)GetProcAddress(hModuleA, "ippiResizeCubicInit_8u");
	pfippiResizeSuperInit_8u = (PFippiResizeSuperInit_8u)GetProcAddress(hModuleA, "ippiResizeSuperInit_8u");
	pfippiResizeCubic_8u_C3R = (PFippiResizeCubic_8u_C3R)GetProcAddress(hModuleA, "ippiResizeCubic_8u_C3R");
	pfippiResizeSuper_8u_C3R = (PFippiResizeSuper_8u_C3R)GetProcAddress(hModuleA, "ippiResizeSuper_8u_C3R");
    pfippiResizeGetBufferSize_8u = (PFippiResizeGetBufferSize_8u)GetProcAddress(hModuleA, "ippiResizeGetBufferSize_8u");
	pfippiResizeGetSrcOffset_8u = (PFippiResizeGetSrcOffset_8u)GetProcAddress(hModuleA, "ippiResizeGetSrcOffset_8u");

	HMODULE hModuleB = GetModuleHandle("ipps.dll");
	
	if (hModuleB == NULL)
	{
		hModuleB = LoadLibrary("ipps.dll");
		if (hModuleB == NULL)
		{
			MessageBox(0, "could not find ipps.dll", "", MB_OK);
			return false;
		}
	}
	pfippsMalloc_8u = (PFippsMalloc_8u)GetProcAddress(hModuleB, "ippsMalloc_8u");
	pfippsFree = (PFippsFree)GetProcAddress(hModuleB, "ippsFree");

	return true;
}

bool ResizeByIntel2019(unsigned char* pSrcPixels, SIZE srcSize, int srcLineLength, RECT srcRoi,
	unsigned char* pDestPixels, int destLineLength, SIZE destSize,
	double xScale, double yScale, int interpolation, int channels)
{
	IppiResizeSpec_32f* pSpec = 0;
	int specSize = 0, initSize = 0, bufSize = 0;
	Ipp8u* pBuffer = 0;
	Ipp8u* pInitBuf = 0;
	IppiPoint dstOffset = { 0, 0 };
	IppStatus status = ippStsNoErr;
	IppiBorderType border = ippBorderRepl;

	IppiSize ippi_srcSize, ippi_destSize;
	ippi_srcSize.width = srcSize.cx;		ippi_srcSize.height = srcSize.cy;
	ippi_destSize.width = destSize.cx;		ippi_destSize.height = destSize.cy;
	IppiRect ipp_srcRoi;
	ipp_srcRoi.x = srcRoi.left;
	ipp_srcRoi.y = srcRoi.top;
	ipp_srcRoi.width = srcRoi.right - srcRoi.left;
	ipp_srcRoi.height = srcRoi.bottom - srcRoi.top;

	IppiInterpolationType ippIterpolation = ippCubic;
	if (interpolation == 8)
		ippIterpolation = ippSuper;

	// Spec and init buffer sizes
	status = pfippiResizeGetSize_8u(ippi_srcSize, ippi_destSize, ippIterpolation, 0, &specSize, &initSize);
	if (status != ippStsNoErr) return status;

	pInitBuf = pfippsMalloc_8u(initSize);
    pSpec = (IppiResizeSpec_32f*)pfippsMalloc_8u(specSize);

	if (pInitBuf == NULL || pSpec == NULL)
	{
		pfippsFree(pInitBuf);
		pfippsFree(pSpec);
		return false;// ippStsNoMemErr;
	}

	// Filter initialization
	switch (ippIterpolation)
	{
	case ippCubic:
		status = pfippiResizeCubicInit_8u(ippi_srcSize, ippi_destSize, 0, 0.5, pSpec, pInitBuf);
		break;

	/*case ippLinear:
		status = ippiResizeLinearInit_8u(ippi_srcSize, ippi_destSize, pSpec);
		break;

	case ippNearest:
		status = ippiResizeNearestInit_8u(srcSize, dstSize, pSpec);
		break;

	case ippLanczos:
		status = ippiResizeLanczosInit_8u(srcSize, dstSize, 3, pSpec, pInitBuf);
		break;*/

	case ippSuper:
		status = pfippiResizeSuperInit_8u(ippi_srcSize, ippi_destSize, pSpec);
		break;

	default:
		status = ippStsInterpolationErr;
		break;
	}

	pfippsFree(pInitBuf);
	if (status != ippStsNoErr)
	{
		pfippsFree(pSpec);
		return status;
	}

	// work buffer size
	status = pfippiResizeGetBufferSize_8u(pSpec, ippi_destSize, channels, &bufSize);
	if (status != ippStsNoErr)
	{
		pfippsFree(pSpec);
        return status;
	}

	pBuffer = pfippsMalloc_8u(bufSize);
	if (pBuffer == NULL)
	{
		pfippsFree(pSpec);
		return false;// ippStsNoMemErr;
	}

	Ipp8u* pSrc = (Ipp8u*)pSrcPixels;
	Ipp8u* pDst = (Ipp8u*)pDestPixels;
	int srcStep = srcLineLength;
	int dstStep = destLineLength;

	int type = channels * 100 + ippIterpolation;

	switch (type)
	{
	//case 101:
	//	status = ippiResizeNearest_8u_C1R(pSrc, srcStep, pDst, dstStep, dstOffset, dstSize, pSpec, pBuffer);
	//	break;

	//case 102:
	//	status = ippiResizeLinear_8u_C1R(pSrc, srcStep, pDst, dstStep, dstOffset, dstSize, ippBorderRepl, 0, pSpec, pBuffer);
	//	break;

	//case 106:
	//	status = ippiResizeCubic_8u_C1R(pSrc, srcStep, pDst, dstStep, dstOffset, dstSize, ippBorderRepl, 0, pSpec, pBuffer);
	//	break;

	//case 108:
	//	status = ippiResizeSuper_8u_C1R(pSrc, srcStep, pDst, dstStep, dstOffset, dstSize, pSpec, pBuffer);
	//	break;

	//case 116:
	//	status = ippiResizeLanczos_8u_C1R(pSrc, srcStep, pDst, dstStep, dstOffset, dstSize, ippBorderRepl, 0, pSpec, pBuffer);
	//	break;

	//case 301:
	//	status = ippiResizeNearest_8u_C3R(pSrc, srcStep, pDst, dstStep, dstOffset, dstSize, pSpec, pBuffer);
	//	break;

	//case 302:
	//	status = ippiResizeLinear_8u_C3R(pSrc, srcStep, pDst, dstStep, dstOffset, dstSize, ippBorderRepl, 0, pSpec, pBuffer);
	//	break;

	case 306:
		status = pfippiResizeCubic_8u_C3R(pSrc, srcStep, pDst, dstStep, dstOffset, ippi_destSize, ippBorderRepl, 0, pSpec, pBuffer);
		break;

	case 308:
		status = pfippiResizeSuper_8u_C3R(pSrc, srcStep, pDst, dstStep, dstOffset, ippi_destSize, pSpec, pBuffer);
		break;

	//case 316:
	//	status = ippiResizeLanczos_8u_C3R(pSrc, srcStep, pDst, dstStep, dstOffset, dstSize, ippBorderRepl, 0, pSpec, pBuffer);
	//	break;

	//case 401:
	//	status = ippiResizeNearest_8u_C4R(pSrc, srcStep, pDst, dstStep, dstOffset, dstSize, pSpec, pBuffer);
	//	break;

	//case 402:
	//	status = ippiResizeLinear_8u_C4R(pSrc, srcStep, pDst, dstStep, dstOffset, dstSize, ippBorderRepl, 0, pSpec, pBuffer);
	//	break;

	//case 406:
	//	status = ippiResizeCubic_8u_C4R(pSrc, srcStep, pDst, dstStep, dstOffset, dstSize, ippBorderRepl, 0, pSpec, pBuffer);
	//	break;

	//case 408:
	//	status = ippiResizeSuper_8u_C4R(pSrc, srcStep, pDst, dstStep, dstOffset, dstSize, pSpec, pBuffer);
	//	break;

	//case 416:
	//	status = ippiResizeLanczos_8u_C4R(pSrc, srcStep, pDst, dstStep, dstOffset, dstSize, ippBorderRepl, 0, pSpec, pBuffer);
	//	break;

	default:
		break;
	}

	pfippsFree(pSpec);
	pfippsFree(pBuffer);
	return status == ippStsNoErr;
}


void ResizeByIntel2019Mod_free(BYTE* toFree)
{
	pfippsFree(toFree);
}

BYTE* ResizeByIntel2019Mod_initSpec(int interpolation, int srcCols, int srcRows,
	int dstCols, int dstRows)
{
	BYTE* retSpec = NULL;
	IppiSize ippi_srcSize, ippi_destSize;
	ippi_srcSize.width = srcCols;
	ippi_srcSize.height = srcRows;
	ippi_destSize.width = dstCols;
	ippi_destSize.height = dstRows;

	int specSize = 0, initSize = 0;
	Ipp8u* pInitBuf = 0;
	IppStatus status = ippStsNoErr;

	IppiInterpolationType ippIterpolation = ippCubic;
	if (interpolation == 8)
		ippIterpolation = ippSuper;

	// Spec and init buffer sizes
	status = pfippiResizeGetSize_8u(ippi_srcSize, ippi_destSize, ippIterpolation, 0, &specSize, &initSize);
	if (status != ippStsNoErr)
	{
		return NULL;
	}

	pInitBuf = pfippsMalloc_8u(initSize);
	retSpec = (BYTE*)pfippsMalloc_8u(specSize);
	if (pInitBuf == NULL || retSpec == NULL)
	{
		pfippsFree(pInitBuf);
		pfippsFree(retSpec);
		return NULL;
	}

	// Filter initialization
	switch (ippIterpolation)
	{
	case ippCubic:
		status = pfippiResizeCubicInit_8u(ippi_srcSize, ippi_destSize, 0, 0.5, (IppiResizeSpec_32f*)(retSpec), pInitBuf);
		break;

	case ippSuper:
		status = pfippiResizeSuperInit_8u(ippi_srcSize, ippi_destSize, (IppiResizeSpec_32f*)(retSpec));
		break;

	default:
		status = ippStsInterpolationErr;
		break;
	}

	pfippsFree(pInitBuf);

	if (status != ippStsNoErr)
	{
		pfippsFree(retSpec);
		return NULL;
	}
	return retSpec;
}

BYTE* ResizeByIntel2019Mod_allocBuffer(int dstCols, int dstColsLast, int dstRows, int dstRowsLast, int channels, int numThreads, BYTE* pSpec, int* bufSize)
{
	int bufSizeCommon = 0;
	int bufSizeLastV = 0;
	int bufSizeLastH = 0;
	int bufSizeLastVH = 0;
	Ipp8u* pBuffer = 0;
	IppiSize ippi_destSize;
	ippi_destSize.width = dstCols;
	ippi_destSize.height = dstRows;
	IppiSize ippi_destSizeLastV;
	ippi_destSizeLastV.width = dstCols;
	ippi_destSizeLastV.height = dstRowsLast;
	IppiSize ippi_destSizeLastH;
	ippi_destSizeLastH.width = dstColsLast;
	ippi_destSizeLastH.height = dstRows;
	IppiSize ippi_destSizeLastVH;
	ippi_destSizeLastVH.width = dstColsLast;
	ippi_destSizeLastVH.height = dstRowsLast;
	
	*bufSize = 0;

	IppStatus status = pfippiResizeGetBufferSize_8u((IppiResizeSpec_32f*)pSpec, ippi_destSize, channels, &bufSizeCommon);
	if (status != ippStsNoErr)
	{
		printf("STATUS 1: %d\n", status);
		return pBuffer;
	}
	status = pfippiResizeGetBufferSize_8u((IppiResizeSpec_32f*)pSpec, ippi_destSizeLastV, channels, &bufSizeLastV);
	if (status != ippStsNoErr)
	{
		printf("STATUS 2: %d\n", status);
		return pBuffer;
	}
	status = pfippiResizeGetBufferSize_8u((IppiResizeSpec_32f*)pSpec, ippi_destSizeLastH, channels, &bufSizeLastH);
	if (status != ippStsNoErr)
	{
		printf("STATUS 3: %d\n", status);
		return pBuffer;
	}
	status = pfippiResizeGetBufferSize_8u((IppiResizeSpec_32f*)pSpec, ippi_destSizeLastVH, channels, &bufSizeLastVH);
	if (status != ippStsNoErr)
	{
		printf("STATUS 4: %d\n", status);
		return pBuffer;
	}

	bufSizeCommon = bufSizeCommon > bufSizeLastH ? bufSizeCommon : bufSizeLastH;
	bufSizeLastVH = bufSizeLastVH > bufSizeLastV ? bufSizeLastVH : bufSizeLastV;
	// 64 bytes alignment
	bufSizeCommon = (bufSizeCommon + 63) & ~63;
	int bufSizeTotal = bufSizeCommon * numThreads;
	if (bufSizeCommon < bufSizeLastVH)
	{
		bufSizeTotal += (bufSizeLastVH - bufSizeCommon);
	}
	*bufSize = bufSizeCommon;
	pBuffer = pfippsMalloc_8u(bufSizeTotal);
	return pBuffer;
}


bool ResizeByIntel2019Mod_resize(unsigned char* pSrcPixels, int srcStride,
	unsigned char* pDestPixels, SIZE dstPosition, SIZE dstSize, int dstStride, double xScale, double yScale, int interpolation, int channels, BYTE* pSpec, BYTE* pBuff)
{
	Ipp8u* pBuffer = (Ipp8u*)pBuff;
	IppiPoint srcOffset = {0, 0};
	IppiPoint dstOffset = { dstPosition.cx, dstPosition.cy };

	IppiSize ippi_destSize;
	ippi_destSize.width = dstSize.cx;
	ippi_destSize.height = dstSize.cy;

	IppStatus status = ippStsNoErr;
	
	
	IppiInterpolationType ippIterpolation = ippCubic;
	if (interpolation == 8)
		ippIterpolation = ippSuper;

	status = pfippiResizeGetSrcOffset_8u((IppiResizeSpec_32f*)pSpec, dstOffset, &srcOffset);

	Ipp8u* pSrc = (Ipp8u*)((Ipp8u*)pSrcPixels + ((size_t)srcOffset.y * srcStride) + (srcOffset.x * channels));
	Ipp8u* pDst = (Ipp8u*)((Ipp8u*)pDestPixels + (dstOffset.y * dstStride) + (dstOffset.x * channels));
	int srcStep = srcStride;
	int dstStep = dstStride;

	int type = channels * 100 + ippIterpolation;
	switch (type)
	{
	case 306:
		status = pfippiResizeCubic_8u_C3R(pSrc, srcStep, pDst, dstStep, dstOffset, ippi_destSize, ippBorderRepl, 0, (IppiResizeSpec_32f*)pSpec, pBuffer);
		break;

	case 308:
		status = pfippiResizeSuper_8u_C3R(pSrc, srcStep, pDst, dstStep, dstOffset, ippi_destSize, (IppiResizeSpec_32f*)pSpec, pBuffer);
		break;
	default:
		break;
	}
	return status == ippStsNoErr;
}